SystemOrganization addCategory: #'System-FilePackage'!

BasicClassOrganizer subclass: #PseudoClassOrganizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PseudoClassOrganizer methodsFor: 'comment accessing' stamp: 'NS 4/6/2004 16:44'!
classComment
	"Answer the comment associated with the object that refers to the receiver."
	classComment == nil ifTrue: [^''].
	^classComment! !

!PseudoClassOrganizer methodsFor: 'comment accessing' stamp: 'NS 4/6/2004 16:44'!
classComment: aChangeRecord
	classComment := aChangeRecord! !

!PseudoClassOrganizer methodsFor: 'accessing' stamp: 'NS 4/6/2004 12:27'!
setDefaultList: aCollection
	super setDefaultList: aCollection.
	self classComment: nil.! !

Object subclass: #FilePackage
	instanceVariableNames: 'fullName sourceSystem classes doIts classOrder'
	classVariableNames: 'LogFileStream'
	poolDictionaries: ''
	category: 'System-FilePackage'!

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'dew 2/13/2004 23:33'!
closeLog
	self logFileStream close.
	LogFileStream := nil.! !

!FilePackage class methodsFor: 'instance creation' stamp: 'dew 10/26/2003 22:08'!
conflictsWithUpdatedMethods: fullName
	| conflicts changeList |
	conflicts := (self fromFileNamed: fullName) conflictsWithUpdatedMethods.
	conflicts isEmpty ifTrue: [^ self].
	changeList := ChangeList new.
	changeList
		changes: conflicts
		file: (FileDirectory default readOnlyFileNamed: fullName) close;
		openAsMorphName: 'Conflicts for ', (FileDirectory localNameFor: fullName)
		multiSelect: true 
! !

!FilePackage class methodsFor: 'reader service' stamp: 'dew 12/16/2002 18:29'!
fileReaderServicesForFile: fullName suffix: suffix

	^(suffix = 'st') | (suffix = 'cs') | (suffix = '*')
		ifTrue: [self services]
		ifFalse: [#()]! !

!FilePackage class methodsFor: 'instance creation' stamp: ''!
fromFileNamed: aName
	^self new fromFileNamed: aName! !

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'dew 2/14/2004 23:03'!
log: aString
	Transcript show: aString.
	self logFileStream nextPutAll: aString.
! !

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'dew 2/14/2004 23:03'!
logCr
	Transcript cr.
	self logFileStream nextPut: Character cr.
! !

!FilePackage class methodsFor: 'conflict checker logging' stamp: 'stephaneducasse 11/1/2005 15:39'!
logFileStream

	LogFileStream ifNil:
		[LogFileStream := FileStream fileNamed: 'ConflictChecker.log'.
		LogFileStream setToEnd].
	^ LogFileStream! !

!FilePackage class methodsFor: 'reader service' stamp: 'dew 12/16/2002 18:29'!
serviceConflictsWithUpdatedMethods
	^ SimpleServiceEntry 
		provider: self 
		label: 'conflicts with updated methods'
		selector: #conflictsWithUpdatedMethods:
		description: 'check for conflicts with more recently updated methods in the image, showing the conflicts in a transcript window'
		buttonLabel: 'conflicts'! !

!FilePackage class methodsFor: 'reader service' stamp: 'dew 12/16/2002 18:28'!
services
	^ Array with: self serviceConflictsWithUpdatedMethods! !

!FilePackage class methodsFor: 'conflict checker preferences' stamp: 'dew 12/22/2002 00:19'!
showIdenticalConflicts
	"(Not implemented yet.  Need to implement versionFromChangeSet first for this to work for the 'false' case.)"
	"Set this to true if we want to show conflicts with methods which have an identical timestamp to the one being checked.  This type of conflict usually just proves that you've already loaded the changeset (or some part of it) in your image."
	^ true
! !

!FilePackage class methodsFor: 'conflict checker preferences' stamp: 'dew 2/13/2004 23:32'!
verboseConflicts
	"Set this to true if we want to list each replacing (potentially conflicting) method being checked."
	^ true! !

!FilePackage methodsFor: 'filein/fileout' stamp: 'alain.plantec 2/6/2009 17:01'!
askForDoits
	| choice choices |
	choices := {'do not process' translated. 'at the beginning' translated. 'at the end' translated}.
	choice := nil.
	[choices includes: choice] whileFalse: [
		choice := UIManager default 
				chooseFrom: choices 
				values: choices 
				title: 'Unprocessed doIts found. When to process those?' translated.
		choice ifNil: [^0]].
	^choices indexOf: choice! !

!FilePackage methodsFor: 'conflict checker' stamp: 'JohanBrichau 10/19/2009 18:16'!
canceledConflictChecking
	^ self canceledConflictChecking: ''! !

!FilePackage methodsFor: 'conflict checker' stamp: 'JohanBrichau 10/19/2009 18:16'!
canceledConflictChecking: aMessage
	self inform: ('Conflict check cancelled: ',aMessage) translated. 
	^ #()! !

!FilePackage methodsFor: 'conflict checker' stamp: 'dew 2/14/2004 00:12'!
checkForMoreRecentUpdateThanChangeSet: updateNumberChangeSet pseudoClass: pseudoClass selector: selector
	"Returns the source code for a conflict if a conflict is found, otherwise returns nil."

	| classOrMeta allChangeSets moreRecentChangeSets conflictingChangeSets changeRecordSource classAndMethodPrintString |

	classAndMethodPrintString := pseudoClass name, (pseudoClass hasMetaclass ifTrue: [' class'] ifFalse: ['']), '>>', selector asString.

	changeRecordSource := pseudoClass sourceCode at: selector.
	changeRecordSource isText
		ifTrue: [changeRecordSource := Text
					fromString: 'method: ', classAndMethodPrintString, ' was removed']
		ifFalse: [changeRecordSource stamp isEmptyOrNil ifTrue:
					[self notify: 'Warning: ', classAndMethodPrintString, ' in ', self packageName, ' has no timestamp/initials!!']].

	pseudoClass exists ifFalse:
		[(self classes at: pseudoClass name) hasDefinition
			ifTrue: [^ nil  "a method was added for a newly defined class; not a conflict"]
			ifFalse: [self class logCr; log: 'CONFLICT found for ', classAndMethodPrintString, '... class ', pseudoClass name asString, ' does not exist in the image and is not defined in the file'.
					^ changeRecordSource]].

	classOrMeta := pseudoClass realClass.

	"Only printout the replacing methods here, but we still check for removed methods too in the rest of this method."
	(self class verboseConflicts and: [classOrMeta includesSelector: selector])
		ifTrue: [self class logCr; log: '...checking ', classOrMeta asString, '>>', selector asString].

	allChangeSets := ChangeSorter allChangeSets.
	moreRecentChangeSets := allChangeSets
				copyFrom: (allChangeSets indexOf: updateNumberChangeSet)
				to: (allChangeSets size).
	conflictingChangeSets := (moreRecentChangeSets select:
		[:cs | (cs atSelector: selector class: classOrMeta) ~~ #none]).
	conflictingChangeSets isEmpty ifTrue: [^ nil].

	self class logCr; log: 'CONFLICT found for ', classAndMethodPrintString,
				(' with newer changeset' asPluralBasedOn: conflictingChangeSets).
	conflictingChangeSets do: [:cs | self class log: ' ', cs name].
	^ changeRecordSource
! !

!FilePackage methodsFor: 'accessing' stamp: ''!
classAt: className
	^self classes at: className! !

!FilePackage methodsFor: 'change record types' stamp: ''!
classComment: chgRec

	(self getClass: chgRec methodClassName) classComment: chgRec! !

!FilePackage methodsFor: 'change record types' stamp: 'al 12/2/2005 13:58'!
classDefinition: string with: chgRec
	| tokens theClass |
	
	self flag: #traits.
		
	tokens := Scanner new scanTokens: string.

	"tokens size = 11 ifFalse:[^doIts add: chgRec]."

	theClass := self getClass: (tokens at: 3).
	theClass definition: string.
	classOrder add: theClass.! !

!FilePackage methodsFor: 'accessing' stamp: ''!
classes
	^classes! !

!FilePackage methodsFor: 'conflict checker' stamp: 'StephaneDucasse 12/5/2009 21:27'!
conflictsWithUpdatedMethods
	"Check this package for conflicts with methods in the image which are in newer updates."

	| localFileName stream updateNumberString updateNumber imageUpdateNumber updateNumberChangeSet conflicts fileStream |

	localFileName := FileDirectory localNameFor: fullName.
	stream := sourceSystem readStream.
	stream upToAll: 'latest update: #'.
	updateNumberString := stream upTo: $].
	stream close.
	
	fileStream := FileStream readOnlyFileNamed: fullName.
	(fileStream contentsOfEntireFile includes: Character linefeed)
		ifTrue: [self notifyWithLabel:  'The changeset file ', localFileName, ' contains linefeeds.  Proceed if...
you know that this is okay (e.g. the file contains raw binary data).'].
	fileStream close.

	updateNumberString isEmpty ifFalse:		"remove prepended junk, if any"
		[updateNumberString := (updateNumberString findTokens: Character space) last].
	updateNumberString asInteger ifNil:
		[updateNumberString := UIManager default
						request: 'Error: ', localFileName, ' has no valid Latest Update number in its header. Please enter the estimated update number (e.g. 4332).' translated.
		(updateNumberString isEmptyOrNil or:	[updateNumberString asInteger isNil])
			ifTrue: [^ self canceledConflictChecking.]].
	updateNumber := updateNumberString asInteger.

	imageUpdateNumber := SystemVersion current highestUpdate.
	updateNumber > imageUpdateNumber ifTrue:
		[(self confirm: 'Warning: The update number for this file (#', updateNumberString, ')
is greater than the highest update number for this image (#', imageUpdateNumber asString, ').
This probably means you need to update your image.
Should we proceed anyway as if the file update number is #', imageUpdateNumber asString, '?')
			ifTrue:
				[updateNumber := imageUpdateNumber.
				updateNumberString := imageUpdateNumber asString]
			ifFalse: [^ self canceledConflictChecking]].

	updateNumberChangeSet := self findUpdateChangeSetMatching: updateNumber.
	updateNumberChangeSet ifNil: [^ self canceledConflictChecking: 'No update set found matching ',updateNumber asString].

	self currentWorld findATranscript: self currentEvent.
	self class logCr; logCr; log: 'Checking ', localFileName, ' (#', updateNumberString, ') for method conflicts with changesets after ', updateNumberChangeSet name, ' ...'.

	conflicts := OrderedCollection new.
	self classes values do: [:pseudoClass |
		(Array with: pseudoClass with: pseudoClass metaClass) do: [:classOrMeta |
			classOrMeta selectorsDo: [:selector | | conflict |
				conflict := self
							checkForMoreRecentUpdateThanChangeSet: updateNumberChangeSet
							pseudoClass: classOrMeta
							selector: selector.
				conflict ifNotNil: [conflicts add: conflict].
			].
		].
	].
	self class logCr; log: conflicts size asString, (' conflict' asPluralBasedOn: conflicts), ' found.'; logCr.
	self class closeLog.
	^ conflicts! !

!FilePackage methodsFor: 'change record types' stamp: ''!
doIt: chgRec
	| string |
	string := chgRec string.
	('*ubclass:*instanceVariableNames:*classVariableNames:*poolDictionaries:*category:*'
		match: string) ifTrue:[^self classDefinition: string with: chgRec].
	('* class*instanceVariableNames:*'
		match: string) ifTrue:[^self metaClassDefinition: string with: chgRec].
	('* removeSelector: *'
		match: string) ifTrue:[^self removedMethod: string with: chgRec].
	('* comment:*'
		match: string) ifTrue:[^self msgClassComment: string with: chgRec].
	('* initialize'
		match: string) ifTrue:[^self]. "Initialization is done based on class>>initialize"
	('''From *'
		match: string) ifTrue:[^self possibleSystemSource: chgRec].
	doIts add: chgRec.! !

!FilePackage methodsFor: 'change record types' stamp: 'avi 1/19/2004 23:47'!
doIts
	^ doIts! !

!FilePackage methodsFor: 'filein/fileout' stamp: 'alain.plantec 2/6/2009 17:02'!
fileIn
	| doitsMark |
	doitsMark := 1.
	doIts isEmpty ifFalse:[doitsMark := self askForDoits].
	doitsMark = 0 ifTrue: [^nil].
	doitsMark = 2 ifTrue:[self fileInDoits].
	classOrder do:[:cls|
		cls fileInDefinition.
	].
	classes do:[:cls|
		Transcript cr; show:'Filing in ', cls name.
		cls fileInMethods.
		cls hasMetaclass ifTrue:[cls metaClass fileInMethods].
	].
	doitsMark = 3 ifTrue:[self fileInDoits].! !

!FilePackage methodsFor: 'filein/fileout' stamp: ''!
fileInDoits
	doIts do:[:chgRec| chgRec fileIn].! !

!FilePackage methodsFor: 'reading' stamp: 'nice 1/5/2010 15:59'!
fileInFrom: aStream
	| changes |
	changes := ChangeSet scanFile: aStream from: 0 to: aStream size.
	aStream close.
	('Processing ', self packageName) 
		displayProgressAt: Sensor cursorPoint
		from: 1
		to: changes size
		during:[:bar| | chgRec |
			1 to: changes size do:[:i|
				bar value: i.
				chgRec := changes at: i.
				self perform: (chgRec type copyWith: $:) asSymbol
with: chgRec.
			].
		].! !

!FilePackage methodsFor: 'filein/fileout' stamp: 'ar 7/17/2005 03:36'!
fileOut
	| fileName stream |
	fileName := UIManager default request: 'Enter the file name' initialAnswer:''.
	stream := FileStream newFileNamed: fileName.
	sourceSystem isEmpty ifFalse:[
		stream nextChunkPut: sourceSystem printString;cr ].
	self fileOutOn: stream.
	stream cr; cr.
	self classes do:[:cls|
		cls needsInitialize ifTrue:[
			stream cr; nextChunkPut: cls name,' initialize']].
	stream cr.
	stream close.

	"DeepCopier new checkVariables."
! !

!FilePackage methodsFor: 'filein/fileout' stamp: ''!
fileOutDoits: aStream
	doIts do:[:chgRec| chgRec fileOutOn: aStream].! !

!FilePackage methodsFor: 'filein/fileout' stamp: 'alain.plantec 2/6/2009 17:02'!
fileOutOn: aStream
	| doitsMark |
	doitsMark := 1.
	doIts isEmpty ifFalse:[doitsMark := self askForDoits].
	doitsMark = 0 ifTrue: [^nil].
	doitsMark = 2 ifTrue:[self fileOutDoits: aStream].
	classOrder do:[:cls|
		cls fileOutDefinitionOn: aStream.
	].
	classes do:[:cls|
		cls fileOutMethodsOn: aStream.
		cls hasMetaclass ifTrue:[cls metaClass fileOutMethodsOn: aStream].
	].
	doitsMark = 3 ifTrue:[self fileOutDoits: aStream].! !

!FilePackage methodsFor: 'conflict checker' stamp: 'dew 10/19/2003 21:29'!
findUpdateChangeSetMatching: updateNumber
	"Find update-changeset beginning with updateNumber, or reasonably close."
	"This is to account for the fact that many changeset files are output from final releases, but may be tested for conflicts in a following alpha image, which will often not include that particular update-changeset from the final release but will contain ones near it.  For example, if the file updateNumber is 5180 (from 3.5 final), but the image has no update-changeset beginning with 5180 because it's a 3.6alpha image (which starts at 5181), it will try up to 5190 and down to 5170 for a close match."
	| updateNumberChangeSet updateNumberToTry |

	updateNumberToTry := updateNumber.
	updateNumberChangeSet := nil.
	[updateNumberChangeSet isNil and: [updateNumberToTry notNil]] whileTrue:
		[updateNumberChangeSet := ChangeSorter allChangeSets
			detect: [:cs | (cs name beginsWith: updateNumberToTry asString)
							and: [(cs name at: (updateNumberToTry asString size + 1)) isDigit not]]
			ifNone: [nil].
		updateNumberToTry >= updateNumber ifTrue:
			[updateNumberToTry < (updateNumber + 10)
				ifTrue: [updateNumberToTry := updateNumberToTry + 1]
				ifFalse: [updateNumberToTry := updateNumber]].
		updateNumberToTry <= updateNumber ifTrue:
			[updateNumberToTry > (updateNumber - 10)
				ifTrue: [updateNumberToTry := updateNumberToTry - 1]
				ifFalse: [updateNumberToTry := nil  "we're done trying"]].
		].

	updateNumberChangeSet ifNil:
		[(self confirm: 'Warning: No changeset beginning with ',
updateNumber asString, ' (within +/- 10) was found in the image.
You must have changesets going back this far in your image
in order to accurately check for conflicts.
Proceed anyway?')
			ifTrue: [updateNumberChangeSet := ChangeSorter allChangeSets first]].

	^ updateNumberChangeSet! !

!FilePackage methodsFor: 'accessing' stamp: 'nk 2/18/2004 18:31'!
fixClassOrder
	"Essentially bubble sort the classOrder so that superclasses appear before subclasses"
	| superClass index subClass superIndex |
	index := 0.
	[index < classOrder size] whileTrue:[
		subClass := classOrder at: (index := index + 1).
		superClass := nil.
		subClass isMeta ifTrue:[
			"Treat non-meta as superclass"
			superClass := self classes at: subClass name ifAbsent:[nil].
		] ifFalse:[
			subClass hasDefinition ifTrue:[
				superClass := self classes 
					at: (Scanner new scanTokens: subClass definition) first ifAbsent:[nil].
				superClass ifNotNil:[superClass hasDefinition ifFalse:[superClass := nil]].
			].
		].
		superClass ifNotNil:[
			superIndex := classOrder indexOf: superClass ifAbsent:[self error:'Where is the class?'].
			superIndex > index ifTrue:[
				"Move superClass before index"
				classOrder remove: superClass.
				classOrder add: superClass before: subClass.
				"Rewind index - we need to check superClass itself"
				index := index - 1.
			].
		].
	].
! !

!FilePackage methodsFor: 'initialize' stamp: 'StephaneDucasse 2/2/2010 21:52'!
fromFileNamed: aName
	
	fullName := aName.
	FileStream 
		readOnlyFileNamed: aName 
		do: [ :aStream |
			aStream setConverterForCode.
			self fileInFrom: aStream] ! !

!FilePackage methodsFor: 'initialize' stamp: 'yo 8/17/2004 09:54'!
fromFileNamed: aName encoding: encodingName
	| stream |
	fullName := aName.
	stream := FileStream readOnlyFileNamed: aName.
	stream converter: (TextConverter newForEncoding: encodingName).
	self fileInFrom: stream.! !

!FilePackage methodsFor: 'accessing' stamp: 'pnm
8/23/2000 17:10'!
fullName: aString
	fullName := aString! !

!FilePackage methodsFor: 'accessing' stamp: ''!
fullPackageName
	^fullName! !

!FilePackage methodsFor: 'private' stamp: ''!
getClass: className
	| pseudoClass |
	(classes includesKey: className) ifTrue:[
		^classes at: className.
	].
	pseudoClass := PseudoClass new.
	pseudoClass name: className.
	classes at: className put: pseudoClass.
	^pseudoClass.! !

!FilePackage methodsFor: 'initialize' stamp: 'alain.plantec 5/28/2009 09:54'!
initialize
	super initialize.
	classes := Dictionary new.
	classOrder := OrderedCollection new.
	sourceSystem := ''.
	doIts := OrderedCollection new.! !

!FilePackage methodsFor: 'private' stamp: ''!
metaClassDefinition: string with: chgRec
	| tokens theClass |
	tokens := Scanner new scanTokens: string.
	theClass := self getClass: (tokens at: 1).
	theClass metaClass definition: string.
	classOrder add: theClass metaClass.! !

!FilePackage methodsFor: 'change record types' stamp: ''!
method: chgRec
	(self getClass: chgRec methodClassName) methodChange: chgRec! !

!FilePackage methodsFor: 'private' stamp: 'ar 4/10/2005 18:46'!
msgClassComment: string with: chgRec
	| tokens theClass |
	tokens := Scanner new scanTokens: string.
	(tokens size = 3 and:[(tokens at: 3) isString]) ifTrue:[
		theClass := self getClass: tokens first.
		^theClass commentString: tokens last].
	(tokens size = 4 and:[(tokens at: 3) asString = 'class' and:[(tokens at: 4) isString]]) ifTrue:[
		theClass := self getClass: tokens first.
		theClass metaClass commentString: tokens last].
! !

!FilePackage methodsFor: 'accessing' stamp: ''!
packageInfo
	^String streamContents:[:s|
		s nextPutAll:'Package: '.
		s nextPutAll: self fullPackageName; cr; cr.
		sourceSystem isEmpty ifFalse:[
			s nextPutAll: sourceSystem; cr; cr].
		doIts isEmpty ifFalse:[
			s nextPutAll:'Unresolvable doIts:'; cr; cr.
			doIts do:[:chgRec|
				s nextPut:$!!; nextPutAll: chgRec string; nextPut: $!!; cr]]].! !

!FilePackage methodsFor: 'accessing' stamp: 'pnm
8/23/2000 17:12'!
packageName
	^FileDirectory localNameFor: self fullPackageName! !

!FilePackage methodsFor: 'private' stamp: 'ar 4/10/2005 18:46'!
possibleSystemSource: chgRec
	| tokens |
	sourceSystem isEmpty ifTrue:[
		tokens := Scanner new scanTokens: chgRec string.
		(tokens size = 1 and:[tokens first isString]) ifTrue:[
			sourceSystem := tokens first.
			^self]].
	doIts add: chgRec.! !

!FilePackage methodsFor: 'change record types' stamp: ''!
preamble: chgRec
	self doIt: chgRec! !

!FilePackage methodsFor: 'accessing' stamp: ''!
removeClass: aPseudoClass
	(self classes removeKey: aPseudoClass name).
	classOrder copy do:[:cls|
		cls name = aPseudoClass name ifTrue:[ classOrder remove: cls].
	].! !

!FilePackage methodsFor: 'private' stamp: ''!
removedMethod: string with: chgRec
	| class tokens |
	tokens := Scanner new scanTokens: string.
	(tokens size = 3 and:[(tokens at: 2) == #removeSelector: ]) ifTrue:[
		class := self getClass: (tokens at: 1).
		^class removeSelector: (tokens at: 3).
	].
	(tokens size = 4 and:[(tokens at: 2) == #class and:[(tokens at: 3) == #removeSelector:]]) ifTrue:[
		class := self getClass: (tokens at: 1).
		^class metaClass removeSelector: (tokens at: 4).
	].
	doIts add: chgRec! !

!FilePackage methodsFor: 'accessing' stamp: ''!
renameClass: aPseudoClass to: newName
	| oldName |
	oldName := aPseudoClass name.
	self classes removeKey: oldName.
	self classes at: newName put: aPseudoClass.
	aPseudoClass renameTo: newName.! !

!FilePackage methodsFor: 'private' stamp: ''!
sampleMethod
"	In an existing method there are always a number of changes.
	Other stuff
		will be deleted
	Or even better,
		some things may be just modified.
"! !

Object subclass: #PseudoClass
	instanceVariableNames: 'name definition organization source metaClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PseudoClass commentStamp: '<historical>' prior: 0!
I provide an inert model of a Class, used by FileContentsBrowser to manipulate filedout code. Instead of a method dictionary or selectors onto CompiledMethods, I have a dictionary ("source") of selectors onto ChangeRecords, which were, in the case of FileContentsBrowser, parsed from a source or change set file.!

!PseudoClass methodsFor: 'methods' stamp: 'sma 6/1/2000 14:54'!
addMethodChange: aChangeRecord
	| selector |
	selector := self parserClass new parseSelector: aChangeRecord string.
	self organization classify: selector under: aChangeRecord category.
	self sourceCodeAt: selector put: aChangeRecord! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 4/29/2004 06:59'!
allCallsOn
	^ (self realClass ifNil: [ ^#() ]) allCallsOn! !

!PseudoClass methodsFor: 'accessing' stamp: 'sma 6/16/1999 22:59'!
allInstVarNames
	^#()! !

!PseudoClass methodsFor: 'private' stamp: 'nk 2/18/2004 18:33'!
allSubclassesWithLevelDo: classAndLevelBlock startingLevel: level
	^ (self realClass ifNil: [ ^self ])  allSubclassesWithLevelDo: classAndLevelBlock startingLevel: level! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 2/18/2004 18:32'!
allSuperclasses
	^ (self realClass ifNil: [ ^#() ]) allSuperclasses! !

!PseudoClass methodsFor: 'testing method dictionary' stamp: 'marcus.denker 11/10/2008 10:04'!
bindingOf: varName
	self exists ifTrue:[
		(self realClass bindingOf: varName) ifNotNil:[:binding| ^binding].
	].
	^Smalltalk bindingOf: varName asSymbol! !

!PseudoClass methodsFor: 'class' stamp: ''!
classComment: aChangeRecord
	self organization classComment: aChangeRecord! !

!PseudoClass methodsFor: 'errors' stamp: ''!
classNotDefined
	^self inform: self name,' is not defined in the system.
You have to define this class first.'.! !

!PseudoClass methodsFor: 'class' stamp: 'di 1/13/1999 12:00'!
classPool 
	self exists ifFalse: [^ nil].
	^ self realClass classPool! !

!PseudoClass methodsFor: 'class' stamp: 'BJP 4/23/2001 13:50'!
comment
	| rStr |
	rStr := self organization commentRemoteStr.
	^rStr isNil
		ifTrue:[self name,' has not been commented in this file']
		ifFalse:[rStr string]! !

!PseudoClass methodsFor: 'class' stamp: ''!
comment: aString
	self commentString: aString.! !

!PseudoClass methodsFor: 'class' stamp: 'ar 2/5/2004 15:18'!
commentString
	^self comment asString! !

!PseudoClass methodsFor: 'class' stamp: ''!
commentString: aString
	self classComment: aString asText. "Just wrap it"! !

!PseudoClass methodsFor: 'accessing' stamp: 'lr 3/14/2010 21:13'!
compilerClass
	^ (Smalltalk globals at: name ifAbsent: [ ^ Compiler ]) compilerClass! !

!PseudoClass methodsFor: 'private' stamp: ''!
confirmRemovalOf: aString
	^self confirm:'Remove ',aString,' ?'! !

!PseudoClass methodsFor: 'class' stamp: 'LC 10/8/2001 04:46'!
definition
	| link linkText defText |
	^definition ifNil:
		[defText := Text fromString: 'There is no class definition for '.
		link := TextLink new.
		linkText := link analyze: self name with: 'Definition'.
		linkText := Text string: (linkText ifNil: ['']) attribute: link.
		defText append: linkText; append: ' in this file'].! !

!PseudoClass methodsFor: 'class' stamp: ''!
definition: aString
	definition := aString! !

!PseudoClass methodsFor: 'private' stamp: ''!
evaluate: aString
	^Compiler evaluate: aString for: nil logged: true! !

!PseudoClass methodsFor: 'testing' stamp: 'lr 3/14/2010 21:13'!
exists
	^ (Smalltalk globals at: self name asSymbol ifAbsent: [ ^ false ]) isKindOf: Behavior! !

!PseudoClass methodsFor: 'filein/fileout' stamp: ''!
fileIn
	"FileIn the receiver"
	self hasDefinition ifTrue:[self fileInDefinition].
	self fileInMethods: self selectors.
	metaClass ifNotNil:[metaClass fileIn].
	self needsInitialize ifTrue:[
		self evaluate: self name,' initialize'.
	].! !

!PseudoClass methodsFor: 'filein/fileout' stamp: ''!
fileInCategory: aCategory
	^self fileInMethods: (self organization listAtCategoryNamed: aCategory)! !

!PseudoClass methodsFor: 'filein/fileout' stamp: 'ar 7/16/2005 14:06'!
fileInDefinition
	self hasDefinition ifFalse:[^self].
	(self makeSureSuperClassExists: (definition copyUpTo: Character space)) ifFalse:[^self].
	self hasDefinition ifTrue:[
		Transcript cr; show:'Defining ', self name.
		self evaluate: self definition].
	self exists ifFalse:[^self].
	self hasComment ifTrue:[self realClass classComment: self comment].! !

!PseudoClass methodsFor: 'filein/fileout' stamp: ''!
fileInMethod: selector
	^self fileInMethods: (Array with: selector)! !

!PseudoClass methodsFor: 'filein/fileout' stamp: ''!
fileInMethods
	^self fileInMethods: self selectors! !

!PseudoClass methodsFor: 'filein/fileout' stamp: 'nice 1/5/2010 15:59'!
fileInMethods: aCollection
	"FileIn all methods with selectors taken from aCollection"
	| theClass |
	self exists ifFalse:[^self classNotDefined].
	theClass := self realClass.
	aCollection do:[:sel| | cat |
		cat := self organization categoryOfElement: sel.
		cat = self removedCategoryName ifFalse:[
			theClass 
				compile: (self sourceCodeAt: sel) 
				classified: cat
				withStamp: (self stampAt: sel)
				notifying: nil.
		].
	].! !

!PseudoClass methodsFor: 'filein/fileout' stamp: 'PeterHugossonMiller 9/3/2009 10:52'!
fileOut
	| internalStream |
	internalStream := (String new: 1000) writeStream.
	self fileOutOn: internalStream.
	self needsInitialize ifTrue:[
		internalStream cr; nextChunkPut: self name,' initialize'.
	].

	FileStream writeSourceCodeFrom: internalStream baseName: self name isSt: true! !

!PseudoClass methodsFor: 'filein/fileout' stamp: 'PeterHugossonMiller 9/3/2009 10:52'!
fileOutCategory: categoryName

	| internalStream |
	internalStream := (String new: 1000) writeStream.
	self fileOutMethods: (self organization listAtCategoryNamed: categoryName)
			on: internalStream.
	FileStream writeSourceCodeFrom: internalStream baseName: (self name, '-', categoryName) isSt: true! !

!PseudoClass methodsFor: 'fileIn/fileOut' stamp: 'mir 9/25/2008 15:04'!
fileOutDefinitionOn: aStream
	self hasDefinition ifFalse:[^self].
	aStream nextChunkPut: self definition; cr.
	self hasComment ifTrue: [
		aStream cr.
		self organization commentRemoteStr fileOutOn: aStream]! !

!PseudoClass methodsFor: 'filein/fileout' stamp: 'PeterHugossonMiller 9/3/2009 10:52'!
fileOutMethod: selector
	| internalStream |

	internalStream := (String new: 1000) writeStream.

	self fileOutMethods: (Array with: selector) on: internalStream.

	FileStream writeSourceCodeFrom: internalStream baseName: (self name , '-' , (selector copyReplaceAll: ':' with: '')) isSt: true! !

!PseudoClass methodsFor: 'filein/fileout' stamp: 'ar 2/7/2004 01:04'!
fileOutMethods: aCollection on: aStream
	"FileOut all methods with selectors taken from aCollection"
	| cat categories |
	categories := Dictionary new.
	aCollection do:[:sel|
		cat := self organization categoryOfElement: sel.
		cat = self removedCategoryName ifFalse:[
			(categories includesKey: cat) 
				ifFalse:[categories at: cat put: Set new].
			(categories at: cat) add: sel].
	].
	categories associationsDo:[:assoc|
		cat := assoc key.
		assoc value do:[:sel|
			aStream cr.
			(self sourceCode at: sel) fileOutOn: aStream.
		].
	].! !

!PseudoClass methodsFor: 'filein/fileout' stamp: ''!
fileOutMethodsOn: aStream
	^self fileOutMethods: self selectors on: aStream.! !

!PseudoClass methodsFor: 'filein/fileout' stamp: ''!
fileOutOn: aStream
	"FileOut the receiver"
	self fileOutDefinitionOn: aStream.
	metaClass ifNotNil:[metaClass fileOutDefinitionOn: aStream].
	self fileOutMethods: self selectors on: aStream.
	metaClass ifNotNil:[metaClass fileOutMethods: metaClass selectors on: aStream].! !

!PseudoClass methodsFor: 'accessing' stamp: ''!
fullName
	^self name! !

!PseudoClass methodsFor: 'testing' stamp: 'marcus.denker 7/29/2009 15:27'!
hasChanges

	self sourceCode isEmpty ifFalse:[^true].
	self organization hasComment ifTrue:[^true].
	definition isNil ifFalse:[^true].
	metaClass isNil ifFalse:[^metaClass hasChanges].
	^false! !

!PseudoClass methodsFor: 'testing' stamp: ''!
hasComment
	^self organization commentRemoteStr notNil! !

!PseudoClass methodsFor: 'testing' stamp: ''!
hasDefinition
	^definition notNil! !

!PseudoClass methodsFor: 'testing' stamp: ''!
hasMetaclass
	^metaClass notNil! !

!PseudoClass methodsFor: 'methods' stamp: 'StephaneDucasse 2/20/2010 21:53'!
includesSelector: aSymbol 
	"Answer whether the message whose selector is the argument is in the 
	method dictionary of the receiver's class."

	^ self exists
		ifTrue: [self realClass includesSelector: aSymbol]
		ifFalse: [false]
		
	"Note that this method is only used to check whether the class should initialized so returning false is better than raising an error"! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 3/9/2004 10:24'!
instVarNames
	^ #()! !

!PseudoClass methodsFor: 'override' stamp: 'nk 2/18/2004 18:30'!
isMeta
	^false! !

!PseudoClass methodsFor: 'printing' stamp: 'sma 6/17/1999 00:00'!
literalScannedAs: scannedLiteral notifying: requestor 
	^ scannedLiteral! !

!PseudoClass methodsFor: 'private' stamp: 'lr 3/14/2010 21:13'!
makeSureClassExists: aString
	| theClass |
	theClass := Smalltalk globals at: aString asSymbol ifAbsent: [ nil ].
	theClass ifNotNil: [ ^ true ].
	^ self confirm: aString , ' does not exist in the system. Use nil instead?'! !

!PseudoClass methodsFor: 'private' stamp: 'lr 3/14/2010 21:13'!
makeSureSuperClassExists: aString
	| theClass |
	theClass := Smalltalk globals at: aString asSymbol ifAbsent: [ nil ].
	theClass ifNotNil: [ ^ true ].
	^ self confirm: 'The super class ' , aString , ' does not exist in the system. Use nil instead?'! !

!PseudoClass methodsFor: 'class' stamp: ''!
metaClass
	^metaClass ifNil:[metaClass := PseudoMetaclass new name: (self name)].! !

!PseudoClass methodsFor: 'methods' stamp: ''!
methodChange: aChangeRecord
	aChangeRecord isMetaClassChange ifTrue:[
		^self metaClass addMethodChange: aChangeRecord.
	] ifFalse:[
		^self addMethodChange: aChangeRecord.
	].
! !

!PseudoClass methodsFor: 'accessing' stamp: ''!
name
	^name! !

!PseudoClass methodsFor: 'accessing' stamp: ''!
name: anObject
	name := anObject! !

!PseudoClass methodsFor: 'testing' stamp: 'lr 3/14/2010 21:13'!
nameExists
	^ Smalltalk globals includesKey: self name asSymbol! !

!PseudoClass methodsFor: 'testing' stamp: 'StephaneDucasse 2/20/2010 21:50'!
needsInitialize
	^self hasMetaclass 
		and: [self metaClass includesSelector: #initialize]! !

!PseudoClass methodsFor: 'accessing' stamp: 'NS 4/6/2004 15:46'!
organization
	organization ifNil: [organization := PseudoClassOrganizer defaultList: SortedCollection new].

	"Making sure that subject is set correctly. It should not be necessary."
	organization setSubject: self.
	^ organization! !

!PseudoClass methodsFor: 'private' stamp: 'ajh 1/21/2003 13:03'!
parserClass

	^ Compiler parserClass! !

!PseudoClass methodsFor: 'accessing' stamp: 'nk 2/23/2005 18:00'!
prettyPrinterClass
	^self class prettyPrinterClass! !

!PseudoClass methodsFor: 'printing' stamp: 'ar 2/5/2004 16:04'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut:$(; print: name; nextPut:$)! !

!PseudoClass methodsFor: 'accessing' stamp: 'lr 3/14/2010 21:13'!
realClass
	^ Smalltalk globals at: self name asSymbol ifAbsent: [  ]! !

!PseudoClass methodsFor: 'removing' stamp: ''!
removeAllUnmodified
	| stClass |
	self exists ifFalse:[^self].
	self removeUnmodifiedMethods: self selectors.
	stClass := self realClass.
	(self hasDefinition and:[stClass definition = self definition]) ifTrue:[definition := nil].
	(self hasComment and:[stClass comment asString = self commentString]) ifTrue:[ self classComment: nil].
	metaClass isNil ifFalse:[metaClass removeAllUnmodified].! !

!PseudoClass methodsFor: 'categories' stamp: ''!
removeCategory: selector
	(self organization listAtCategoryNamed: selector) do:[:sel|
		self organization removeElement: sel.
		self sourceCode removeKey: sel.
	].
	self organization removeCategory: selector.! !

!PseudoClass methodsFor: 'methods' stamp: ''!
removeMethod: selector
	self organization removeElement: selector.
	self sourceCode removeKey: selector.
! !

!PseudoClass methodsFor: 'methods' stamp: ''!
removeSelector: aSelector
	| catName |
	catName := self removedCategoryName.
	self organization addCategory: catName before: self organization categories first.
	self organization classify: aSelector under: catName.
	self sourceCodeAt: aSelector put:'methodWasRemoved' asText.! !

!PseudoClass methodsFor: 'removing' stamp: ''!
removeUnmodifiedMethods: aCollection
	| stClass |
	self exists ifFalse:[^self].
	stClass := self realClass.
	aCollection do:[:sel|
		(self sourceCodeAt: sel) = (stClass sourceCodeAt: sel ifAbsent:['']) asString ifTrue:[
			self removeMethod: sel.
		].
	].
	self organization removeEmptyCategories.! !

!PseudoClass methodsFor: 'categories' stamp: ''!
removedCategoryName
	^'*** removed methods ***' asSymbol! !

!PseudoClass methodsFor: 'class' stamp: 'nk 2/18/2004 18:30'!
renameTo: aString

	self hasDefinition ifTrue:[
		self isMeta ifTrue:[
			self definition: (self definition
				copyReplaceAll: name,' class'
				with: aString, ' class').
		] ifFalse:[
			self definition: (self definition 
					copyReplaceAll:'ubclass: #',name
					with:'ubclass: #', aString)]].
	name := aString.
	metaClass ifNotNil:[metaClass renameTo: aString].! !

!PseudoClass methodsFor: 'methods' stamp: ''!
selectors
	^self sourceCode keys! !

!PseudoClass methodsFor: 'class' stamp: 'di 1/13/1999 12:00'!
sharedPools 
	self exists ifFalse: [^ nil].
	^ self realClass sharedPools! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCode
	^source ifNil:[source := Dictionary new]! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCodeAt: sel
	^(self sourceCode at: sel) string! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCodeAt: sel put: object
	self sourceCode at: sel put: object! !

!PseudoClass methodsFor: 'methods' stamp: ''!
sourceCodeTemplate
	^''! !

!PseudoClass methodsFor: 'methods' stamp: 'sw 6/10/2003 17:31'!
stampAt: selector
	"Answer the authoring time-stamp of the change"

	| code |
	^ ((code := self sourceCode at: selector) isKindOf: ChangeRecord)
		ifTrue:
			[code stamp]
		ifFalse:
			[code string]! !

!PseudoClass methodsFor: 'accessing' stamp: 'NorbertHartl 6/20/2008 21:25'!
theMetaClass
	^ self metaClass! !

!PseudoClass methodsFor: 'accessing' stamp: 'wod 5/19/1998 17:42'!
theNonMetaClass
	"Sent to a class or metaclass, always return the class"

	^self! !

!PseudoClass methodsFor: 'categories' stamp: ''!
whichCategoryIncludesSelector: aSelector 
	"Answer the category of the argument, aSelector, in the organization of 
	the receiver, or answer nil if the receiver does not inlcude this selector."

	^ self organization categoryOfElement: aSelector! !

PseudoClass subclass: #PseudoMetaclass
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PseudoMetaclass methodsFor: 'accessing' stamp: ''!
fullName
	^self name,' class'! !

!PseudoMetaclass methodsFor: 'testing' stamp: 'nk 2/18/2004 18:30'!
isMeta
	^true! !

!PseudoMetaclass methodsFor: 'accessing' stamp: ''!
realClass
	^super realClass class! !

!PseudoMetaclass methodsFor: 'accessing' stamp: 'FBS 3/4/2004 14:17'!
theNonMetaClass
	"Sent to a class or metaclass, always return the class"

	^self realClass theNonMetaClass! !

Object subclass: #TextDiffBuilder
	instanceVariableNames: 'realSrc realDst srcMap dstMap srcLines dstLines srcPos dstPos added removed shifted runs matches multipleMatches patchSequence'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

TextDiffBuilder subclass: #ClassDiffBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!ClassDiffBuilder methodsFor: 'printing' stamp: 'nice 1/5/2010 15:59'!
printPatchSequence: ps on: aStream 
	
	ps do: [:assoc | | type line | 
			type := assoc key.
			line := assoc value.
			aStream
				withAttributes: (self attributesOf: type)
				do: [aStream nextPutAll: line]]! !

!ClassDiffBuilder methodsFor: 'initialize' stamp: 'PeterHugossonMiller 9/3/2009 00:56'!
split: aString 
	| lines in out c |
	lines := OrderedCollection new.
	in := aString readStream.
	out := String new writeStream.
	[ in atEnd ] whileFalse: 
		[ (c := in next) isSeparator 
			ifTrue: 
				[ out nextPut: c.
				lines add: out contents.
				out reset ]
			ifFalse: [ out nextPut: c ] ].
	out position = 0 ifFalse: [ lines add: out contents ].
	^ lines! !

TextDiffBuilder subclass: #CodeDiffBuilder
	instanceVariableNames: 'class'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!CodeDiffBuilder commentStamp: '<historical>' prior: 0!
I am a differencer that compares source in tokens tokenised by a parser.  I consider comments significant, but consider sequences of whitespace equivalent.  Depending on the definition of WhitespaceForCodeDiff>>at: sequences of whitespace containing carriage-returns may be considered different to sequences of whitespace lacking carriage-returns (which may result in better-formatted diffs).!

TextDiffBuilder subclass: #PrettyTextDiffBuilder
	instanceVariableNames: 'sourceClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!PrettyTextDiffBuilder class methodsFor: 'instance creation' stamp: 'nk 10/29/2000 12:35'!
from: srcString to: dstString inClass: srcClass 
	^ (self new sourceClass: srcClass) from: srcString to: dstString
		! !

!PrettyTextDiffBuilder methodsFor: 'initialize' stamp: 'nk 10/29/2000 12:16'!
sourceClass: aClass
	sourceClass := aClass.! !

!PrettyTextDiffBuilder methodsFor: 'initialize' stamp: 'alain.plantec 5/18/2009 15:59'!
split: aString 
	| formatted trimmed |
	trimmed := aString asString withBlanksTrimmed.
	trimmed isEmpty ifTrue: [ ^super split: '' ].
	formatted := [ sourceClass prettyPrinterClass
				format: trimmed
				in: sourceClass
				notifying: nil] on: Error do: [ :ex | trimmed ].
	^ super split: formatted! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: ''!
buildDisplayPatchFrom: srcString to: dstString
	^(self from: srcString to: dstString) buildDisplayPatch! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: 'AlainPlantec 12/1/2009 22:05'!
buildDisplayPatchFrom: srcString to: dstString inClass: srcClass 
	^ ((srcClass notNil and: [CodeHolder diffsWithPrettyPrint])
		ifTrue: [PrettyTextDiffBuilder
				from: srcString
				to: dstString
				inClass: srcClass]
		ifFalse: [self from: srcString to: dstString]) buildDisplayPatch! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: 'sw 5/19/2001 10:52'!
buildDisplayPatchFrom: srcString to: dstString inClass: srcClass prettyDiffs: prettyBoolean
	"Build a display patch for mapping via diffs from srcString to dstString in the given class.  If prettyBoolean is true, do the diffing for pretty-printed forms"

	^ ((srcClass notNil and: [prettyBoolean])
		ifTrue: [PrettyTextDiffBuilder
				from: srcString
				to: dstString
				inClass: srcClass]
		ifFalse: [self from: srcString to: dstString]) buildDisplayPatch! !

!TextDiffBuilder class methodsFor: 'instance creation' stamp: ''!
from: srcString to: dstString
	^self new from: srcString to: dstString! !

!TextDiffBuilder methodsFor: 'private' stamp: 'DamienPollet 11/24/2008 22:18'!
attributesOf: type
	"Private.
	Answer the TextAttributes that are used to display text of the given type."

	^type caseOf: {
		[#insert] -> [ {TextEmphasis bold. TextColor color: Color green muchDarker} ].
		[#remove] -> [ {TextEmphasis bold. TextColor color: Color red darker} ].
	} otherwise: [ {TextEmphasis normal} ].
! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: ''!
buildDisplayPatch
	^Text streamContents:[:stream|
		self printPatchSequence: self buildPatchSequence on: stream.
	]! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'RAA 5/2/2001 23:35'!
buildPatchSequence
	"@@ TODO: Das funktioniert noch nicht fr n-m matches"
	matches := TwoLevelDictionary new.
	self buildReferenceMap.
	runs := self processDiagonals.
	self validateRuns: runs.
	"There may be things which have just been moved around. Find those."
	shifted := self detectShiftedRuns.
	self processShiftedRuns.
	"Now generate a patch sequence"
	patchSequence := self generatePatchSequence.
	^patchSequence! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'ar 11/20/1998 16:57'!
buildReferenceMap
	dstLines doWithIndex:[:line :index|
		(srcPos at: line ifAbsent:[#()]) 
			do:[:index2| matches at: index@index2 put: line]
	].
	srcLines doWithIndex:[:line :index|
		(dstPos at: line ifAbsent:[#()]) 
			do:[:index2| matches at: index2@index put: line]
	].
! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: ''!
collectRunFrom: todo startingWith: startIndex into: run
	| next start |
	start := startIndex.
	self remove: start from: todo.
	run add: (matches at: start).
	"Search downwards"
	next := start.
	[next := next + (1@1).
	todo includes: next] whileTrue:[
		run addLast: (matches at: next).
		self remove: next from: todo].
	"Search upwards"
	next := start.
	[next := next - (1@1).
	todo includes: next] whileTrue:[
		run addFirst: (matches at: next).
		self remove: next from: todo.
		start := next. "To use the first index"
	].
	^start! !

!TextDiffBuilder methodsFor: 'initialize' stamp: 'nk 10/29/2000 12:15'!
destString: aString 
	realDst := self split: aString asString.
	dstLines := OrderedCollection new.
	dstMap := OrderedCollection new.
	realDst
		doWithIndex: [:line :realIndex | 
			dstLines
				add: (self formatLine: line).
			dstMap add: realIndex].
	dstPos := PluggableDictionary new: dstLines size.
	dstPos hashBlock: self stringHashBlock.
	dstLines
		doWithIndex: [:line :index | (dstPos includesKey: line)
				ifTrue: [(dstPos at: line)
						add: index.
					multipleMatches := true]
				ifFalse: [dstPos
						at: line
						put: (OrderedCollection with: index)]]! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: ''!
detectShiftedRuns
	| sortedRuns lastY run shiftedRuns |
	runs size < 2 ifTrue: [^ nil].
	shiftedRuns := OrderedCollection new.
	sortedRuns := SortedCollection sortBlock: [:a1 :a2 | a1 key x < a2 key x].
	runs associationsDo: [:assoc | sortedRuns add: assoc].
	lastY := sortedRuns first key y.
	2 to: sortedRuns size do:[:i | 
		run := sortedRuns at: i.
		run key y > lastY
			ifTrue: [lastY := run key y]
			ifFalse: [shiftedRuns add: run]].
	^ shiftedRuns! !

!TextDiffBuilder methodsFor: 'initialize' stamp: 'nk 1/7/2004 09:24'!
formatLine: aString
	^aString copyWithout: Character lf! !

!TextDiffBuilder methodsFor: 'initialize' stamp: ''!
from: sourceString to: destString
	self sourceString: sourceString.
	self destString: destString.! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: ''!
generatePatchSequence
	| ps |
	ps := OrderedCollection new: srcLines size.
	srcLines size timesRepeat:[ps add: nil].
	self incorporateMatchesInto: ps.
	self incorporateRemovalsInto: ps.
	self incorporateAddsInto: ps.
	^ps! !

!TextDiffBuilder methodsFor: 'testing' stamp: ''!
hasMultipleMatches
	^multipleMatches == true! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'di 3/15/1999 14:01'!
incorporateAddsInto: aPatchSequence
	"Incorporate adds"
	| lastMatch lastIndex index |
	added ifNil:[^self].
	added := added sortBy:[:a1 :a2| a1 key < a2 key].
	lastMatch := 1.
	lastIndex := 0.
	1 to: added size do:[:i|
		index := (added at: i) key.
		[index > lastMatch] whileTrue:[
			[lastIndex := lastIndex + 1.
			(aPatchSequence at: lastIndex) key == #match] whileFalse.
			lastMatch := lastMatch + 1.
		].
		aPatchSequence add: #insert->(added at: i) value afterIndex: lastIndex.
		lastIndex := lastIndex + 1.
		lastMatch := lastMatch + 1.
	].! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'nice 1/5/2010 15:59'!
incorporateMatchesInto: aPatchSequence
	"Incorporate matches"
	
	runs associationsDo:[:assoc| | index |
		index := assoc key y.
		assoc value do:[:line|
			self assert:[(aPatchSequence at: index) isNil].
			aPatchSequence at: index put: (#match -> line).
			index := index + 1.
		].
	].
! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'nice 1/5/2010 15:59'!
incorporateRemovalsInto: aPatchSequence
	"Incorporate removals"
	
	removed ifNil:[^self].
	removed do:[:assoc| | index |
		index := assoc key.
		self assert:[(aPatchSequence at: index) isNil].
		aPatchSequence at: index put: #remove -> assoc value.
	].
! !

!TextDiffBuilder methodsFor: 'private' stamp: 'stephane.ducasse 4/13/2009 21:12'!
pointEqualBlock
	^[ :a :b | a x = b x and: [a y = b y]] ! !

!TextDiffBuilder methodsFor: 'private' stamp: 'stephane.ducasse 4/13/2009 21:12'!
pointHashBlock
	^[:pt| (pt x bitShift: 12) + pt y] ! !

!TextDiffBuilder methodsFor: 'printing' stamp: 'nk 4/24/2004 08:48'!
printPatchSequence: seq on: aStream 
	seq do: 
		[:assoc | 
		aStream
			withAttributes: (self attributesOf: assoc key)
			do: [aStream nextPutAll: assoc value; cr]]! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'RAA 5/2/2001 23:41'!
processDiagonals

	^self processDiagonalsFrom: matches twoLevelKeys
! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'RAA 5/2/2001 23:17'!
processDiagonalsFrom: todoList
	| runList start run todo |
	todo := todoList copy.
	runList := PluggableDictionary new.
	runList hashBlock: self pointHashBlock.
	runList equalBlock: self pointEqualBlock.
	[todo isEmpty] whileFalse:[
		start := todo detect:[:any| true].
		run := OrderedCollection new.
		start := self 
					collectRunFrom: todo 
					startingWith: start 
					into: run.
		runList at: start put: run.
	].
	"If we have multiple matches we might have chosen a bad sequence.
	There we redo the whole thing recursively"
	self hasMultipleMatches  ifFalse:[^runList].
	runList size < 2 ifTrue:[^runList].

	run := nil.
	start := 0.
	runList associationsDo:[:assoc|
		(run isNil or:[assoc value size > run size]) ifTrue:[
			run := assoc value.
			start := assoc key]].
	"Now found the longest run"
	run := OrderedCollection new.
	start := self
				collectRunFrom: todoList
				startingWith: start
				into: run.
	"Find the diagonals in the remaining set"
	runList := self processDiagonalsFrom: todoList.
	runList at: start put: run.
	^runList! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'nice 1/5/2010 15:59'!
processShiftedRuns
	
	shifted isNil ifTrue:[^self].
	shifted do:[:assoc| | key |
		key := assoc key.
		assoc value doWithIndex:[:line :idx|
			removed add: (key y + idx - 1) -> line.
			added add: (key x + idx - 1) -> line].
		runs removeKey: assoc key.
	].
! !

!TextDiffBuilder methodsFor: 'private' stamp: 'RAA 5/2/2001 23:28'!
remove: pointKey from: aSet

	self hasMultipleMatches ifFalse:[^aSet remove: pointKey].
	aSet removeAllXAndY: pointKey.
! !

!TextDiffBuilder methodsFor: 'initialize' stamp: 'nk 10/29/2000 12:15'!
sourceString: aString 
	realSrc := self split: aString asString.
	srcLines := OrderedCollection new.
	srcMap := OrderedCollection new.
	realSrc
		doWithIndex: [:line :realIndex | 
			srcLines
				add: (self formatLine: line).
			srcMap add: realIndex].
	srcPos := PluggableDictionary new: srcLines size.
	srcPos hashBlock: self stringHashBlock.
	srcLines
		doWithIndex: [:line :index | (srcPos includesKey: line)
				ifTrue: [(srcPos at: line)
						add: index.
					multipleMatches := true]
				ifFalse: [srcPos
						at: line
						put: (OrderedCollection with: index)]]! !

!TextDiffBuilder methodsFor: 'initialize' stamp: ''!
split: aString
	^self split: aString by: self splitCharacter! !

!TextDiffBuilder methodsFor: 'private' stamp: 'ar 11/20/1998 17:26'!
split: aString by: splitChar
	| lines index nextIndex |
	lines := OrderedCollection new.
	index := 1.
	[index <= aString size] whileTrue:[
		nextIndex := aString 
						indexOf: splitChar 
						startingAt: index 
						ifAbsent:[aString size+1].
		lines add: (aString copyFrom: index to: nextIndex-1).
		index := nextIndex+1].
	^lines! !

!TextDiffBuilder methodsFor: 'private' stamp: ''!
splitCharacter
	^Character cr! !

!TextDiffBuilder methodsFor: 'private' stamp: 'nice 1/5/2010 15:59'!
stringHashBlock
	"Return a block for use in string hashing"
	
	^[:string| | stringSize | 
		stringSize := string size.
		stringSize = 0 
			ifTrue:[0]
			ifFalse:[ stringSize < 3 
				ifTrue:[(string at: 1) asInteger +
						((string at: string size) asInteger bitShift: 8)]
				ifFalse:[	(string at: 1) asInteger +
						((string at: stringSize // 3 + 1) asInteger bitShift: 4) +
						((string at: stringSize // 2 + 1) asInteger bitShift: 8) +
						((string at: stringSize * 2 // 3 + 1) asInteger bitShift: 12) +
						((string at: stringSize) asInteger bitShift: 16)]]] ! !

!TextDiffBuilder methodsFor: 'creating patches' stamp: 'nice 1/5/2010 15:59'!
validateRuns: runList
	| srcPosCopy dstPosCopy |
	srcPosCopy := srcPos copy.
	srcPosCopy associationsDo:[:assoc| assoc value: assoc value asSet].
	dstPosCopy := dstPos copy.
	dstPosCopy associationsDo:[:assoc| assoc value: assoc value asSet].
	runList associationsDo:[:assoc| | dstIndex lines srcIndex |
		srcIndex := assoc key y.
		dstIndex := assoc key x.
		lines := assoc value.
		lines do:[:string|
			(srcPosCopy at: string) remove: srcIndex.
			(dstPosCopy at: string) remove: dstIndex.
			srcIndex := srcIndex + 1.
			dstIndex := dstIndex + 1.
		].
	].
	removed := OrderedCollection new.
	srcPosCopy associationsDo:[:assoc|
		assoc value do:[:index| removed add: (index -> assoc key)].
	].
	removed := removed sortBy:[:a1 :a2| a1 key < a2 key].
	added := OrderedCollection new.
	dstPosCopy associationsDo:[:assoc|
		assoc value do:[:index| added add: (index -> assoc key)].
	].
	added := added sortBy:[:a1 :a2| a1 key < a2 key].
! !

Object subclass: #TwoLevelDictionary
	instanceVariableNames: 'firstLevel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!TwoLevelDictionary commentStamp: '<historical>' prior: 0!
A simple dictionary for the use of the TextDiffBuilder. Keys are presumed to be Points and a significant speed advantage is gained by using a dictionary of dictionaries. The first is keyed by the x-values and the second by the y-values. Only the minimum necessary protocol is implemented.!

!TwoLevelDictionary methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:42'!
at: aPoint

	^(firstLevel at: aPoint x ifAbsent: [^nil]) at: aPoint y ifAbsent: [^nil]
! !

!TwoLevelDictionary methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:37'!
at: aPoint put: anObject

	(firstLevel at: aPoint x ifAbsentPut: [Dictionary new]) at: aPoint y put: anObject
! !

!TwoLevelDictionary methodsFor: 'as yet unclassified' stamp: 'alain.plantec 5/28/2009 11:14'!
initialize

	super initialize.
	firstLevel := Dictionary new.! !

!TwoLevelDictionary methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:40'!
keysDo: aBlock

	firstLevel keysAndValuesDo: [ :x :v |
		v keysDo: [ :y | aBlock value: x@y]
	].! !

!TwoLevelDictionary methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:40'!
twoLevelKeys

	| twoLevelSet |

	twoLevelSet := TwoLevelSet new.
	self keysDo: [ :each | twoLevelSet add: each].
	^twoLevelSet
! !

Object subclass: #TwoLevelSet
	instanceVariableNames: 'firstLevel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-FilePackage'!

!TwoLevelSet commentStamp: '<historical>' prior: 0!
A simple set for the use of the TextDiffBuilder. Elements are presumed to be Points and a significant speed advantage is gained by using a dictionary of sets. The first is keyed by the x-values and the second contains the y-values. Only the minimum necessary protocol is implemented.!

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:18'!
add: aPoint

	(firstLevel at: aPoint x ifAbsentPut: [Set new]) add: aPoint y
! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:18'!
copy

	| answer |

	answer := self class new initialize.
	self do: [ :each |
		answer add: each
	].
	^answer! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/3/2001 09:26'!
detect: aBlock

	firstLevel keysAndValuesDo: [ :x :v |
		v do: [ :y | (aBlock value: x@y) ifTrue: [^x@y]]
	].
	^nil! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:15'!
do: aBlock

	firstLevel keysAndValuesDo: [ :x :v |
		v do: [ :y | aBlock value: x@y]
	].! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:19'!
includes: aPoint

	^(firstLevel at: aPoint x ifAbsent: [^false]) includes: aPoint y! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'alain.plantec 5/28/2009 11:15'!
initialize

	super initialize.
	firstLevel := Dictionary new.! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:13'!
isEmpty

	^firstLevel isEmpty! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:22'!
remove: aPoint

	| lev2 |

	lev2 := firstLevel at: aPoint x ifAbsent: [^self].
	lev2 remove: aPoint y ifAbsent: [].
	lev2 isEmpty ifTrue: [firstLevel removeKey: aPoint x].

! !

!TwoLevelSet methodsFor: 'as yet unclassified' stamp: 'RAA 5/2/2001 23:28'!
removeAllXAndY: aPoint

	| deletes |

	deletes := OrderedCollection new.
	firstLevel removeKey: aPoint x ifAbsent: [].
	firstLevel keysAndValuesDo: [ :x :lev2 |
		lev2 remove: aPoint y ifAbsent: [].
		lev2 isEmpty ifTrue: [deletes add: x].
	].
	deletes do: [ :each | firstLevel removeKey: each ifAbsent: []].! !
